#include "uogetopt.h"
#include "alloc.h"
#include "env.h"
#include "str.h"
#include "bailout.h"
#include "error.h"
#include "r822.h"
#include "pathexec.h"
#include "pathexec_argv.h"
#include "stralloc.h"
#include "readwrite.h"
#include "seek.h"
#include "mail_in.h"
#include "attributes.h"
#include "ch_dir.h"
#include "get_pw_name.h"

#define PROGRAM "usendmail"

static void die_seek(void) attribute_noreturn;
static void die_newaliases(void) attribute_noreturn;
static void die_daemonmode(void) attribute_noreturn;
static void die_queuemode(void) attribute_noreturn;
static void die_mode(const char *x) attribute_noreturn;
static void smtpd(void) attribute_noreturn;
static void mailq(void) attribute_noreturn;
static void runx(const char *) attribute_noreturn;
static void qstat(void) attribute_noreturn;
static void tcpto(void) attribute_noreturn;
static void tcpok(void) attribute_noreturn;

static void die_seek(void)
{xbailout(111,errno,"failed to read in stdin",0,0,0); }
static void die_newaliases(void)
{xbailout(2,0,"please use fastforward/newaliases instead.",0,0,0);}
static void die_daemonmode(void)
{xbailout(2,0,"daemonmode not implemented. start qmail instead.",0,0,0);}
static void die_queuemode(void)
{xbailout(2,0,"queuemode not implemented. send an alarm to qmail-send.",0,0,0);}
static void die_mode(const char *x) 
{xbailout(2,0,"mode ",x," not implemented.",0);}

static void runx(const char *p)
{
	union {const char *cc; char *c;} dq;
	char *av[2];
	dq.cc=p;
	av[0]=dq.c;
	av[1]=0;
	pathexec(av);
	xbailout(111,errno, "failed to run ",p,0,0);
}

static void smtpd(void)
{
	if (!env_get("PROTO")) {
		if (!pathexec_env("RELAYCLIENT","")) oom();
		if (!pathexec_env("DATABYTES","0")) oom();
		if (!pathexec_env("PROTO","TCP")) oom();
		if (!pathexec_env("TCPLOCALIP","127.0.0.1")) oom();
		if (!pathexec_env("TCPLOCALHOST","localhost")) oom();
		if (!pathexec_env("TCPREMOTEIP","127.0.0.1")) oom();
		if (!pathexec_env("TCPREMOTEHOST","localhost")) oom();
		if (!pathexec_env("TCPREMOTEINFO","usendmail-bs")) oom();
	}
	runx("bin/qmail-smtpd");
}

static void mailq(void) { runx("bin/qmail-qread"); }
static void qstat(void) { runx("bin/qmail-qstat"); }
static void tcpto(void) { runx("bin/qmail-tcpto"); }
static void tcpok(void) { runx("bin/qmail-tcpok"); }

const char auto_qmail[]="/var/qmail\0PATCHSPACE------------------------END";
static int flag_dummyi;
static int flag_debug;
static char *flag_env_debug;
static int flag_address_debug;
static unsigned long flag_dummyul;
static char *flag_dummys;
static int flag_rec_in_header;
static char *flag_env_sender;
static char *flag_fullname;
static char *flag_b;
static int flag_newaliases;

/* unbelievable: there are still option letters left unused. You would 
 * believe that after many, many (almost 20!) years of careless increase
 * of the bloat that resource would be lost, wouldn't you?
 * Don't worry, the sendmail people just overused -O, -o and -b.
 */
static uogetopt_t opts[]={
/* -A xxx     use alternate .cf - no way, no use, no brain. */
{'B',0,UOGO_STRING,&flag_dummys,0,"Set body type (ignored).",0},
/* -C xxx     set configuration file. Of course not. */
{'b',0,UOGO_STRING,&flag_b,0,
"Select operation mode.\n"
"-bm (mail delivery), -bp (print queue) and\n"
"-bs (smtpd mode) are implemented.","MODE"},
{'c',0,UOGO_FLAG|UOGO_HIDDEN,&flag_dummys,0,
"Connect to non-local mailers (ignored).",0},
/* -d is debugging */
{'e',0,UOGO_STRING,&flag_dummys,0,
"Error mode (ignored).\n"
"Only one error mode is implemented:\n"
"Failures to queue the mail are reported immediately\n"
"with an exit code, other errors are reported\n"
"through a bounce message.","MODE"},
{'E',0,UOGO_STRING|UOGO_HIDDEN|UOGO_OPTARG,&flag_dummys,0,
	"Code conversion. Ignored.\n"
	"Sony NEWS OS uses this?",0},
{'F',0,UOGO_STRING,&flag_fullname,0,"Set fullname of sender.\n"
"This information is used to build the from header.","NAME"},
{'f',0,UOGO_STRING,&flag_env_sender,0,"Set envelope sender.\n"
"This is the address error messages are sent to.","SENDER"},
{'G',0,UOGO_FLAG,&flag_dummyi,0,"Gateway submission (ignored).\n", 0},
{'h',0,UOGO_ULONG,&flag_dummyul,0,"Set hopcount (ignored).\n"
"Which insane mind would want to defeat loop\n"
"detection locally only? Remove the Received-Headers\n"
"instead.",0},
{'i',0,UOGO_FLAG,&flag_dummyi,0,"Ignore dot lines (ignored).\n"
"Ignore dots alone on lines by themselves in\n"
"incoming messages. This is completely stupid.",0},
{'I',0,UOGO_FLAG,&flag_newaliases,1,"Initialize alias DBM file (forbidden).\n"
 "Use newaliases from the fastforward package instead.",0},
{'J',0,UOGO_STRING|UOGO_HIDDEN|UOGO_OPTARG,&flag_dummys,0,
	"Code conversion. Ignored.\n"
	"Sony NEWS OS uses this?",0},
{'L',0,UOGO_STRING|UOGO_HIDDEN|UOGO_OPTARG,&flag_dummys,0,
	"Logging label. Ignored.\n","LABEL"},
{'m',0,UOGO_FLAG,&flag_dummyi,0,"Send to me too (ignored).",0},
/* -M: define macro */
{'N',0,UOGO_STRING,&flag_dummys,0,"Set DSN mode (ignored).\n"
"Set delivery status notification mode.\n","MODE"},
{'n',0,UOGO_FLAG,&flag_dummys,0,"Don't do aliasing (ignored).\n"
PROGRAM " doesn't do aliasing anyway.\n"
"This is the job of qmail.",0},
{'O',0,UOGO_STRING,&flag_dummys,0,"Set long option (ignored).",0},
{'o',0,UOGO_STRING,&flag_dummys,0,"Set short option (ignored).",0},
{'p',0,UOGO_STRING,&flag_dummys,0,"Set protocol (ignored).",0},
/* -q is neither ignorable nor implementable */
/* -Q is neither ignorable nor implementable */
/* "full", "hdrs" */
{'R',0,UOGO_STRING,&flag_dummyul,0,"Set bounce size (ignored).",0},
{'r',0,UOGO_STRING|UOGO_HIDDEN,&flag_env_sender,0,"Set envelope sender.",0},
{'s',0,UOGO_FLAG,&flag_dummys,0,"Save from lines in header (ignored).",0},
{'t',0,UOGO_FLAG,&flag_rec_in_header,1,
	"Recipients in header.\n"
	"Read recipients from the message header, too.",0},
{'T',0,UOGO_FLAG,&flag_dummyul,0,"Set timeout interval (ignored).",0},
{'U',0,UOGO_FLAG|UOGO_HIDDEN,&flag_dummyi,0,"Initial user submission (ignored).",0},
{'v',0,UOGO_FLAG|UOGO_HIDDEN,&flag_dummyi,0,"Verbose (ignored).",0},
/* DSN ENVID: set "original" envelope id */
{'V',0,UOGO_FLAG|UOGO_HIDDEN,&flag_dummyi,0,"set 'original' envelope id */ (ignored).",0},
/* -X: traffic log file. Huh? Brain dump? */
{'X',0,UOGO_STRING|UOGO_HIDDEN,&flag_dummys,0,"Set traffic log file (ignored).",0},
{'x',0,UOGO_FLAG|UOGO_HIDDEN,&flag_dummyi,0,"Unknown option (ignored).",0},
{0,"debug",UOGO_FLAG,&flag_debug,1,"debug " PROGRAM ".\n"
"Causes qmail-inject to print the message.\n"
"The message will not be sent.",0},
{0,"address-debug",UOGO_FLAG,&flag_address_debug,1,"debug address handling.\n"
"Print the arguments that would be passed to\n"
"qmail-inject. The message will not be sent.",0},
{0,"environment-debug",UOGO_STRING,&flag_env_debug,0,"debug environment.\n"
"Print the environment.\n"
"Do not call qmail-inject.","PRINTENV"},
{0,0,0,0,0,0,0}
};


static int got_mailsuser;
static int got_mailshost;
static stralloc suser;
static stralloc shost;
static void
handle_str(const char *s)
{
	unsigned int i;
	i=str_chr(s,'@');
	if (!i) return; /* can't use this: @host */
	if (s[i] && s[i+1] && !got_mailshost) {
		if (!stralloc_copys(&shost,s+i+1)) oom();
		if (!stralloc_0(&shost)) oom();
		if (!pathexec_env("QMAILSHOST",shost.s)) oom();
	}
	if (!got_mailsuser) {
		if (!stralloc_copyb(&suser,s,i)) oom();
		if (!stralloc_0(&suser)) oom();
		if (!pathexec_env("QMAILSUSER",suser.s)) oom();
	}
	got_mailsuser=1;
	got_mailshost=1; /* means empty host, yes */
}

static void 
make_env_sender(const char *from, const char *sender)
{
	static stralloc k,add;
	unsigned int i;
	if (flag_env_sender)
		return; /* will send with -f */
	if (got_mailsuser)
		return;
	if (sender) {
		if (!stralloc_copys(&add,sender)) oom();
		switch (r822_address(&k, &add, 0)) {
		case R822_OK:
			for (i=0;i<k.len;i+=str_len(k.s+i)+1)
				if (k.s[i]=='A') {
					handle_str(k.s+i+1);
					if (got_mailsuser)
						return;
					break;
				}
			break;
		case R822_ERR_NOMEM: oom();
		default:
			break;
		}
	}

	if (!stralloc_copys(&add,from)) oom();
	switch (r822_address(&k, &add, 0)) {
	case R822_OK:
		for (i=0;i<k.len;i+=str_len(k.s+i)+1)
			if (k.s[i]=='A') {
				handle_str(k.s+i+1);
				if (got_mailsuser)
					return;
				break;
			}
		break;
	case R822_ERR_NOMEM: oom();
	default:
		break;
	}
}

/* make sure qmail-inject has a valid user */
static void
defaultfrom(void)
{
  const char *p;

  if (flag_env_sender) {
    static stralloc sa;
    unsigned int o;
    if (env_get("QMAILUSER")) return;
    if (!stralloc_copys(&sa,flag_env_sender)) oom();
    if (!stralloc_0(&sa)) oom();
    o=str_chr(sa.s,'@');
    if (sa.s[o]) {
      sa.s[o]=0;
      if (sa.s[o+1])
	if (!env_get("QMAILHOST"))
	  if (!pathexec_env("QMAILHOST",sa.s+o+1)) oom();
    }
    if (!pathexec_env("QMAILUSER",sa.s)) oom();
    stralloc_free(&sa);
    return;
  }
  p=env_get("QMAILUSER");
  if (!p)
    p=env_get("MAILUSER");
  if (!p)
    p=env_get("USER");
  if (!p)
    p=env_get("LOGNAME");
  if (!p) {
    p=get_pw_name();
    if (p && !pathexec_env("USER",p)) oom();
  }
}


int 
main(int argc,char **argv)
{
	int j;
	const char *from=0;
	const char *sender=0;
	char **nav;

	flag_bailout_fatal_begin=3;
	bailout_progname(argv[0]);

	if (ch_dir(auto_qmail) == -1)
		xbailout(111,errno, "failed to switch to qmail home directory",0,0,0);

	uogetopt (flag_bailout_log_name, PACKAGE, VERSION,
		&argc, argv, uogetopt_out,
		"usage: usendmail [options] [address ...]\n",
		opts,
		"\nReport bugs to <uwe-usendmail@bulkmail.ohse.de>");

	if (flag_newaliases) die_newaliases();
	if (flag_b) {
		if (flag_b[1]) die_mode(flag_b);
		switch (flag_b[0]) {
		case 'h': tcpto(); /* hoststat */
		case 'H': tcpok(); /* purge hoststat */
		case 'm': break; /* deliver mail in the usual way */
		case 'p': mailq(); /* print a listing of the queue */
		case 'P': qstat(); /* print the number of entries in the queue */
		case 's': smtpd();
		case 'd': die_daemonmode(); /* run as daemon */
		case 'D': die_daemonmode(); /* run as daemon in foreground */
		case 'i': die_newaliases();
		case 'q': die_queuemode(); /* run the queue */
		default: die_mode(flag_b);
			break;
		}
	}
	if (flag_fullname) 
		if (!pathexec_env("MAILNAME",flag_fullname))
			oom();

	if (str_equal(flag_bailout_log_name,"mailq")) mailq();
	if (str_equal(flag_bailout_log_name,"hoststat")) tcpto();
	if (str_equal(flag_bailout_log_name,"purgestat")) tcpok();
	if (str_equal(flag_bailout_log_name,"newaliases")) die_newaliases();
	/* note: getroot does more than just an smtp dialog if started with the 
	 * name "smtpd". */
	if (str_equal(flag_bailout_log_name,"smtpd")) die_daemonmode();

	if (env_get("QMAILSHOST"))
		got_mailshost=1;
	if (env_get("QMAILUSER") || env_get("MAILUSER") || env_get("USER")
		|| env_get("LOGNAME") || env_get("QMAILSUSER"))
		got_mailsuser=1;
	if (!got_mailsuser)
		got_mailshost=0;
	if (-1!=seek_begin(0))  {
		mail_parse_header(0);
		if (-1==seek_begin(0)) 
			die_seek();
		from=mail_header("from");
		sender=mail_header("sender");
	}
	if (from) {
		const char *s=env_get("QMAILINJECT");
		if (s) {
			unsigned int i;
			i=str_chr(s,'f');
			if (s[i]) {
				/* tell's qmail-inject to delete incoming from field.
				 * This program will not honor from then.
				 */
				from=0;
			}
		}
	}
	if (from || sender)
		make_env_sender(from,sender);
	if (!from)
		defaultfrom();

	if (!pathexec_argv_add("bin/qmail-inject")) oom();
	if (flag_rec_in_header) {
		if (!pathexec_argv_add("-H")) oom();
	} else {
		if (!pathexec_argv_add("-a")) oom();
	}
	if (flag_debug)
		if (!pathexec_argv_add("-n")) oom();
	if (flag_env_sender) {
		if (!pathexec_argv_add("-f")) oom();
		if (!pathexec_argv_add(flag_env_sender)) oom();
	}
	if (!pathexec_argv_add("--")) oom();
	/* "uwe-1,2"@ohse.de,uwe-3@nrw.net
	 *   S: "uwe-1,2"@ohse.de
	 *   S: uwe-3@nrw.net
	 * uwe-"1,2@ohse.de"@ohse.de
	 *   S: uwe-"1,2@ohse.de"@ohse.de
	 *     btw: this is wrong. "uwe-" is a quoted-string, "..." is an atom,
	 *     both types are "words".
	 *     the local part has to be in this format:  word *("." word)
	 *     Sendmail accepts an invalid email address here.
	 * usendmail could recognize that one, but i decided to be 
	 * compatible to the getroot-program.
	 */
	for (j = 1;j < argc;++j) {
		unsigned int i;
		stralloc sa=STRALLOC_INIT;
		stralloc k=STRALLOC_INIT;
		if (!stralloc_copys(&sa,argv[j])) oom();
		switch (r822_address(&k, &sa, 1)) { /* 1: strict */
		case R822_OK:
			for (i=0;i<k.len;i+=str_len(k.s+i)+1)
				if (k.s[i]=='A' && k.s[i+1])
					if (!pathexec_argv_add(k.s+i+1)) oom();
			break;
		case R822_ERR_NOMEM: oom();
		default:
			if (!pathexec_argv_add(argv[j])) oom();
			break;
		}
	}
	if (flag_env_debug) {
		pathexec_argv_reset();
		if (!pathexec_argv_add(flag_env_debug)) oom();
	} 
	nav=pathexec_argv_get();
	if (flag_address_debug) {
		unsigned int i;
		for (i=0;nav[i];i++) {
			write(1,nav[i],str_len(nav[i])); 
			write(1,"\n",1);
		}
		return 0;
	}
	pathexec(nav);
	xbailout(111,errno,PROGRAM
		": fatal: failed to start ",nav[0],0,0);
}
